# coding=utf-8


class TdxConnectionError(Exception):
    """
    当连接服务器出错的时候，会抛出的异常
    """
    pass

class TdxFunctionCallError(Exception):
    """
    当行数调用出错的时候
    """
    def __init__(self, *args, **kwargs):
        super(TdxFunctionCallError, self).__init__(*args, **kwargs)
        self.original_exception = None







#coding: utf-8
import click

from concurrent.futures import ThreadPoolExecutor
from pytdx.hq import TdxHq_API
import datetime
import time
from pytdx.util.best_ip import select_best_ip

GET_QUOTES_PER_GROUPS=80

@click.command()
@click.argument("ips", nargs=-1)
def main(ips):

    if len(ips) == 0:
        best_ip = select_best_ip()
        ips = [best_ip]
        print("Using default ip: {}".format(best_ip))

    def single_client_benchmark(ip):

        def _log(msg):
            click.echo("HQ_BENCHMARK: [{:15s}] {} ".format(ip, datetime.datetime.now()) + msg)

        def _grouped_list(stocks):
            return [stocks[i:i + GET_QUOTES_PER_GROUPS] for i in range(0, len(stocks), GET_QUOTES_PER_GROUPS)]

        _log("start benchmark")

        total_time = connecting_time = get_security_count_time = get_security_list_time = get_security_quotes_time = num = 0

        start_time = time.time()
        last_time = start_time

        try:
            api = TdxHq_API(multithread=True)

            port = 7709

            if ":" in ip:
                ip, port = ip.split(':')
                port = int(port)

            with api.connect(ip=ip, port=port):
                _log("connected")
                cur_time = time.time()
                connecting_time = cur_time - last_time
                last_time = cur_time
                _log("connecting time is {}".format(connecting_time))

                num = api.get_security_count(0)
                _log("all shenzhen market stock count is {}".format(num))

                cur_time = time.time()
                get_security_count_time = cur_time - last_time
                last_time = cur_time
                _log("get_security_count_time is {}".format(get_security_count_time))

                all = []
                for i in range((num // 1000) + 1):
                    offset = i * 1000
                    section = api.get_security_list(0, offset)
                    all = all + section

                cur_time = time.time()
                get_security_list_time = cur_time - last_time
                last_time = cur_time

                _log("get_security_list_time is {}".format(get_security_list_time))

                codes = [one['code'] for one in all]

                results = []
                for stocks in _grouped_list(codes):
                    req_list = [(0, code) for code in stocks]
                    one_results = api.get_security_quotes(req_list)
                    results = results + one_results

                cur_time = time.time()
                get_security_quotes_time = cur_time - last_time
                last_time = cur_time
                _log("get_security_quotes_time is {}".format(get_security_quotes_time))

                total_time = last_time - start_time

                _log("total_time is {}".format(total_time))

            _log("end benchmark")
        except Exception as e:
            _log("hit exception " + str(e))

        return {
            "ip": ip,
            "total_time": total_time,
            "connecting_time": connecting_time,
            "get_security_count_time": get_security_count_time,
            "get_security_list_time": get_security_list_time,
            "get_security_quotes_time": get_security_quotes_time,
            "security_count": num
        }

    with ThreadPoolExecutor(max_workers=len(ips)) as executor:
        results = executor.map(single_client_benchmark, ips)



        rows = []
        rows.append(("IP", "Total", "Connecting", "Get Count", "Get List", "Get Quotes"))
        for result in results:
            rows.append(
                [result['ip'],

                 "{:0.6f}".format(result['total_time']),
                 "{:0.6f}".format(result['connecting_time']),

                 "{:0.6f} ({})".format(result['get_security_count_time'], result['security_count']),
                 "{:0.6f}".format(result['get_security_list_time']),

                 "{:0.6f}".format(result['get_security_quotes_time'])]
            )

        print("=" * 40)
        print_table(rows)



# helper function from http://blog.paphus.com/blog/2012/09/04/simple-ascii-tables-in-python/
def print_table(lines, separate_head=True):
  """Prints a formatted table given a 2 dimensional array"""
  #Count the column width
  widths = []
  for line in lines:
      for i,size in enumerate([len(x) for x in line]):
          while i >= len(widths):
              widths.append(0)
          if size > widths[i]:
              widths[i] = size

  #Generate the format string to pad the columns
  print_string = ""
  for i,width in enumerate(widths):
      print_string += "{" + str(i) + ":" + str(width) + "} | "
  if (len(print_string) == 0):
      return
  print_string = print_string[:-3]

  #Print the actual data
  for i,line in enumerate(lines):
      print(print_string.format(*line))
      if (i == 0 and separate_head):
          print("-"*(sum(widths)+3*(len(widths)-1)))


if __name__ == '__main__':
    main()


from pytdx.reader.daily_bar_reader import TdxDailyBarReader, TdxFileNotFoundException, TdxNotAssignVipdocPathException
from pytdx.reader.min_bar_reader import TdxMinBarReader
from pytdx.reader.lc_min_bar_reader import TdxLCMinBarReader
from pytdx.reader.exhq_daily_bar_reader import TdxExHqDailyBarReader
from pytdx.reader.gbbq_reader import GbbqReader
from pytdx.reader.block_reader import BlockReader
from pytdx.reader.block_reader import CustomerBlockReader

__all__ = [
    'TdxDailyBarReader',
    'TdxFileNotFoundException',
    'TdxNotAssignVipdocPathException',
    'TdxMinBarReader',
    'TdxLCMinBarReader',
    'TdxExHqDailyBarReader',
    'GbbqReader',
    'BlockReader',
    'CustomerBlockReader',
]


#coding=utf-8
from __future__ import unicode_literals, division
import struct


class TdxFileNotFoundException(Exception):
    pass

class TdxNotAssignVipdocPathException(Exception):
    pass


class BaseReader(object):

    def unpack_records(self, format, data):
        record_struct = struct.Struct(format)
        return (record_struct.unpack_from(data, offset)
                for offset in range(0, len(data), record_struct.size))

    def get_df(self, code_or_file, exchange=None):
        raise NotImplementedError('not yet')


# coding=utf-8

from pytdx.parser.base import BaseParser
from pytdx.helper import get_datetime, get_volume, get_price
from collections import OrderedDict
import struct
import six

"""
param: category=9, market=1, stockcode=000001, start=0, count=10


In [101]: l[12:19]
Out[101]: bytearray(b'xD\x9eI\xbe\xf7\nR')

should be :

1296527
149215477760.00000


bytearray(b'\xa8\xa9\xa4I\x10R\rR') = 1348917 151741792256.00000

"""



class GetIndexBarsCmd(BaseParser):

    def setParams(self, category, market, code, start, count):
        if type(code) is six.text_type:
            code = code.encode("utf-8")

        self.category = category

        values = (
            0x10c,
            0x01016408,
            0x1c,
            0x1c,
            0x052d,
            market,
            code,
            category,
            1,
            start,
            count,
            0, 0, 0  # I + I +  H total 10 zero
        )

        pkg = struct.pack("<HIHHHH6sHHHHIIH", *values)
        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        pos = 0

        (ret_count,) = struct.unpack("<H", body_buf[0: 2])
        pos += 2

        klines = []

        pre_diff_base = 0
        for i in range(ret_count):
            year, month, day, hour, minute, pos = get_datetime(self.category, body_buf, pos)

            price_open_diff, pos = get_price(body_buf, pos)
            price_close_diff, pos = get_price(body_buf, pos)

            price_high_diff, pos = get_price(body_buf, pos)
            price_low_diff, pos = get_price(body_buf, pos)

            (vol_raw,) = struct.unpack("<I", body_buf[pos: pos + 4])
            vol = get_volume(vol_raw)

            pos += 4
            (dbvol_raw,) = struct.unpack("<I", body_buf[pos: pos + 4])
            dbvol = get_volume(dbvol_raw)
            pos += 4

            (up_count, down_count) = struct.unpack("<HH", body_buf[pos: pos + 4])
            pos += 4

            open = self._cal_price1000(price_open_diff, pre_diff_base)

            price_open_diff = price_open_diff + pre_diff_base

            close = self._cal_price1000(price_open_diff, price_close_diff)
            high = self._cal_price1000(price_open_diff, price_high_diff)
            low = self._cal_price1000(price_open_diff, price_low_diff)

            pre_diff_base = price_open_diff + price_close_diff

            #### 为了避免python处理浮点数的时候，浮点数运算不精确问题，这里引入了多余的代码

            kline = OrderedDict([
                ("open", open),
                ("close", close),
                ("high", high),
                ("low", low),
                ("vol", vol),
                ("amount", dbvol),
                ("year", year),
                ("month", month),
                ("day", day),
                ("hour", hour),
                ("minute", minute),
                ("datetime", "%d-%02d-%02d %02d:%02d" % (year, month, day, hour, minute)),
                ("up_count", up_count),
                ("down_count", down_count)
            ])
            klines.append(kline)
        return klines

    def _cal_price1000(self, base_p, diff):
        return float(base_p + diff)/1000


# coding=utf-8

from pytdx.parser.base import BaseParser
from pytdx.helper import get_datetime, get_volume, get_price
from collections import OrderedDict
import struct
import six


class GetCompanyInfoContent(BaseParser):

    def setParams(self, market, code, filename, start, length):
        if type(code) is six.text_type:
            code = code.encode("utf-8")

        if type(filename) is six.text_type:
            filename = filename.encode("utf-8")

        if len(filename) != 80:
            filename = filename.ljust(80, b'\x00')


        pkg = bytearray.fromhex(u'0c 07 10 9c 00 01 68 00 68 00 d0 02')
        pkg.extend(struct.pack(u"<H6sH80sIII", market, code, 0, filename, start, length, 0))
        self.send_pkg = pkg

    def parseResponse(self, body_buf):
        pos = 0
        _, length = struct.unpack(u'<10sH', body_buf[:12])
        pos += 12
        content = body_buf[pos: pos+length]
        return content.decode("gbk")


# coding:utf-8
from .trade_date import trade_date_sse

import datetime


def get_real_trade_date(date, towards):
    """
    获取真实的交易日期,其中,第三个参数towards是表示向前/向后推
    towards=1 日期向后迭代
    towards=-1 日期向前迭代
    @yutiansut
    """
    if towards == 1:
        while date not in trade_date_sse:
            date = str(datetime.datetime.strptime(
                date, '%Y-%m-%d') + datetime.timedelta(days=1))[0:10]
        else:
            return date
    elif towards == -1:
        while date not in trade_date_sse:
            date = str(datetime.datetime.strptime(
                date, '%Y-%m-%d') - datetime.timedelta(days=1))[0:10]
        else:
            return date



# coding=utf-8

#
# Just for practising
#


import os
import socket
import sys
import pandas as pd

if __name__ == '__main__':
    sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from pytdx.log import DEBUG, log
from pytdx.parser.ex_setup_commands import ExSetupCmd1
from pytdx.parser.ex_get_markets import GetMarkets
from pytdx.parser.ex_get_instrument_count import GetInstrumentCount
from pytdx.parser.ex_get_instrument_quote import GetInstrumentQuote
from pytdx.parser.ex_get_minute_time_data import GetMinuteTimeData
from pytdx.parser.ex_get_history_minute_time_data import GetHistoryMinuteTimeData
from pytdx.parser.ex_get_transaction_data import GetTransactionData
from pytdx.parser.ex_get_history_transaction_data import GetHistoryTransactionData
from pytdx.parser.ex_get_instrument_bars import GetInstrumentBars
from pytdx.parser.ex_get_instrument_info import GetInstrumentInfo
from pytdx.parser.ex_get_history_instrument_bars_range import GetHistoryInstrumentBarsRange
from pytdx.parser.ex_get_instrument_quote_list import GetInstrumentQuoteList


from pytdx.params import TDXParams

import threading
import datetime
from pytdx.base_socket_client import BaseSocketClient, update_last_ack_time


"""
In [7]: 0x7e
Out[7]: 126

In [5]: len(body)
Out[5]: 8066

In [6]: len(body)/126
Out[6]: 64.01587301587301

In [7]: len(body)%126
Out[7]: 2

In [8]: (len(body)-2)/126
Out[8]: 64.0
"""


class TdxExHq_API(BaseSocketClient):

    def setup(self):
        ExSetupCmd1(self.client).call_api()

    # API LIST

    @update_last_ack_time
    def get_markets(self):
        cmd = GetMarkets(self.client)
        return cmd.call_api()

    @update_last_ack_time
    def get_instrument_count(self):
        cmd = GetInstrumentCount(self.client)
        return cmd.call_api()

    @update_last_ack_time
    def get_instrument_quote(self, market, code):
        cmd = GetInstrumentQuote(self.client)
        cmd.setParams(market, code)
        return cmd.call_api()

    @update_last_ack_time
    def get_instrument_bars(self, category, market, code, start=0, count=700):
        cmd = GetInstrumentBars(self.client)
        cmd.setParams(category, market, code, start=start, count=count)
        return cmd.call_api()

    @update_last_ack_time
    def get_minute_time_data(self, market, code):
        cmd = GetMinuteTimeData(self.client)
        cmd.setParams(market, code)
        return cmd.call_api()

    @update_last_ack_time
    def get_history_minute_time_data(self, market, code, date):
        cmd = GetHistoryMinuteTimeData(self.client)
        cmd.setParams(market, code, date=date)
        return cmd.call_api()

    @update_last_ack_time
    def get_transaction_data(self, market, code, start=0, count=1800):
        cmd = GetTransactionData(self.client)
        cmd.setParams(market, code, start=start, count=count)
        return cmd.call_api()

    @update_last_ack_time
    def get_history_transaction_data(self, market, code, date, start=0, count=1800):
        cmd = GetHistoryTransactionData(self.client)
        cmd.setParams(market, code, date, start=start, count=count)
        return cmd.call_api()

    @update_last_ack_time
    def get_history_instrument_bars_range(self, market, code, start, end):
        cmd = GetHistoryInstrumentBarsRange(self.client)
        cmd.setParams(market, code, start, end)
        return cmd.call_api()

    @update_last_ack_time
    def get_instrument_info(self, start, count=100):
        cmd = GetInstrumentInfo(self.client)
        cmd.setParams(start, count)
        return cmd.call_api()

    @update_last_ack_time
    def get_instrument_quote_list(self, market, category, start=0, count=80):
        cmd = GetInstrumentQuoteList(self.client)
        cmd.setParams(market, category, start, count)
        return cmd.call_api()

    def do_heartbeat(self):
        self.get_instrument_count()


if __name__ == '__main__':
    import pprint

    api = TdxExHq_API()
    with api.connect('121.14.110.210', 7727):
        # log.info("获取市场代码")
        # pprint.pprint(api.to_df(api.get_markets()))
        # log.info("查询市场中商品数量")
        # pprint.pprint(api.get_instrument_count())
        # log.info("查询五档行情")
        #pprint.pprint(api.to_df(api.get_instrument_quote(47, "IF1709")))
        #pprint.pprint(api.get_instrument_quote(8, "10000889"))
        #pprint.pprint(api.get_instrument_quote(31, "00020"))
        # log.info("查询分时行情")
        #api.get_minute_time_data(47, "IFL0")
        #pprint.pprint(api.to_df(api.get_minute_time_data(47, "IFL0")))
        #pprint.pprint(api.to_df(api.get_minute_time_data(8, "10000889")).tail())
        #pprint.pprint(api.get_minute_time_data(31, "00020"))
        log.info("查询历史分时行情")
        pprint.pprint(api.to_df(api.get_history_minute_time_data(31, "00020", 20170811)).tail())
        log.info("查询分时成交")
        pprint.pprint(api.to_df(api.get_transaction_data(31, "00020")).tail())

        log.info("查询历史分时成交")
        pprint.pprint(api.to_df(api.get_history_transaction_data(31, "00020", 20170811)).tail())
        #data = api.get_history_minute_time_data(47, 'IFL0', 20170811)
        # pprint.pprint(data)

        # log.info("查询k线")
        #pprint.pprint(api.to_df(api.get_instrument_bars(TDXParams.KLINE_TYPE_DAILY, 8, "10000843")))
        #pprint.pprint(api.to_df(api.get_instrument_bars(TDXParams.KLINE_TYPE_DAILY, 31, "00700")))
        # log.info("查询代码列表")
        #pprint.pprint(api.to_df(api.get_instrument_info(10000, 98)))



# coding=utf-8

from threading import Thread
import random
from pytdx.log import DEBUG, log
import time

# 参考 :https://stackoverflow.com/questions/6524459/stopping-a-thread-after-a-certain-amount-of-time


DEFAULT_HEARTBEAT_INTERVAL = 10.0 # 10秒一个heartbeat

class HqHeartBeatThread(Thread):

    def __init__(self, api, stop_event, heartbeat_interval=DEFAULT_HEARTBEAT_INTERVAL):
        self.api = api
        self.client = api.client
        self.stop_event = stop_event
        self.heartbeat_interval = heartbeat_interval
        super(HqHeartBeatThread, self).__init__()

    def run(self):
        while not self.stop_event.is_set():
            self.stop_event.wait(self.heartbeat_interval)
            if self.client and (time.time() - self.api.last_ack_time > self.heartbeat_interval):
                try:
                    # 发送一个获取股票数量的包作为心跳包
                    self.api.do_heartbeat()
                except Exception as e:
                    log.debug(str(e))


